"""
    BoozerFromFlux()(x::FluxCoordinates, vmec::VmecSurface)
    BoozerFromFlux()(x::AbstractArray{FluxCoordinates}, vmec::VmecSurface)
    BoozerFromFlux()(x::FluxCoordinates, vmec::VmecSurface, wmn::Array{SurfaceFourierData}, boozerG, boozerI)
    BoozerFromFlux()(x::AbstractArray{FluxCoordinates}, vmec::VmecSurface, wmn::Array{SurfaceFourierData}, boozerG, boozerI)

Compute the transformation from flux coordinates to Boozer coordinates using
data from `vmec`, a `VmecSurface` object.  The terms `wmn`, `boozerG` and
`boozerI` need only be computed once per `VmecSurface`.

See also [`periodic_boozer_function`](@ref), [`SurfaceFourierData`](@ref)
"""
function (::BoozerFromFlux)(x::FluxCoordinates,
                            vmec::VmecSurface{T},
                            wmn::AbstractArray{SurfaceFourierData{T}},
                            boozerG::T,
                            boozerI::T;
                           ) where {T <: AbstractFloat}
  w = inverseTransform(x,wmn)
  λ = inverseTransform(x,vmec.λmn)
  jacFac = 1.0/(boozerG + vmec.iota[1]*boozerI)
  p = (w + boozerI*λ)*jacFac
  ψ = x.ψ
  χ = x.θ + λ + vmec.iota[1]*p
  ϕ = x.ζ + p
  return BoozerCoordinates(ψ,χ,ϕ)
end

function (::BoozerFromFlux)(x::FluxCoordinates,
                            vmec::VmecSurface{T};
                           ) where {T <: AbstractFloat}
  wmn, boozerG, boozerI = periodic_boozer_function(vmec)
  return BoozerFromFlux()(x, vmec, wmn, boozerG, boozerI)
end

function (::BoozerFromFlux)(x::AbstractArray{FluxCoordinates},
                            vmec::VmecSurface{T};
                           ) where {T <: AbstractFloat}
  ψ = first(x).ψ
  all(p -> p.ψ == ψ, x) || "Flux surface labels must be identical for each point"
  wmn, boozerG, boozerI = periodic_boozer_function(vmec)
  return BoozerFromFlux()(x, vmec, wmn, boozerG, boozerI)
end

function (::BoozerFromFlux)(x::StructArray{FluxCoordinates{T, T}, N, C, I},
                            vmec::VmecSurface{T},
                            wmn::AbstractArray{SurfaceFourierData{T}},
                            boozerG::T,
                            boozerI::T;
                           ) where {T <: AbstractFloat, N, C, I}
  res = StructArray{BoozerCoordinates{T,T}}(undef, size(x))
  @batch minbatch=16 for i in eachindex(x, res)
    res[i] = BoozerFromFlux()(x[i], vmec, wmn, boozerG, boozerI)
  end
  return res
end


"""
    boozer_jacobian(::BoozerFromFlux, v::FluxCoordinates, vmec::VmecSurface)
    boozer_jacobian(::BoozerFromFlux, v::AbstractArray{FluxCoordinates}, vmec::VmecSurface)
    boozer_jacobian(::BoozerFromFlux, v::FluxCoordinates, vmec::VmecSurface, wmn::Array{SurfaceFourierData}, boozerG, boozerI)
    boozer_jacobian(::BoozerFromFlux, v::AbstractArray{FluxCoordinates}, vmec::VmecSurface, wmn::Array{SurfaceFourierData}, boozerG, boozerI)

Compute the Jacobian of the coordinate transformation to Boozer coordinates from flux coordinates.

See also: [`periodic_boozer_function`](@ref)
"""
function boozer_jacobian(::BoozerFromFlux,
                         v::FluxCoordinates,
                         vmec::VmecSurface{T},
                         wmn::AbstractVector{SurfaceFourierData{T}},
                         boozerG::T,
                         boozerI::T;
                        ) where {T <: AbstractFloat}
  jacFac = 1.0/(boozerG + vmec.iota[1]*boozerI)
  dλdθ = inverseTransform(v,vmec.λmn;deriv=:dθ)
  dλdζ = inverseTransform(v,vmec.λmn;deriv=:dζ)
  dpdθ = (inverseTransform(v,wmn;deriv=:dθ) - boozerI*dλdθ)*jacFac
  dpdζ = (inverseTransform(v,wmn;deriv=:dζ) - boozerI*dλdζ)*jacFac
  return (1+dλdθ)*(1+dpdζ) + (vmec.iota[1] - dλdζ)*dpdθ
end

function boozer_jacobian(::BoozerFromFlux,
                         v::FluxCoordinates,
                         vmec::VmecSurface{T};
                        ) where {T <: AbstractFloat}
  wmn, boozerG, boozerI = periodic_boozer_function(vmec)
  return boozer_jacobian(BoozerFromFlux(), v, vmec, wmn, boozerG, boozerI)
end

function boozer_jacobian(::BoozerFromFlux,
                         v::AbstractArray{FluxCoordinates},
                         vmec::VmecSurface{T};
                        ) where {T <: AbstractFloat}
  ψ = first(v).ψ
  all(p -> p.ψ == ψ, v) || "Flux surface labels must be identical for each point"
  wmn, boozerG, boozerI = periodic_boozer_function(vmec)
  return boozer_jacobian(BoozerFromFlux(), v, vmec, wmn, boozerG, boozerI)
end

function boozer_jacobian(::BoozerFromFlux,
                         v::StructArray{FluxCoordinates{T, T}, N, C, I},
                         vmec::VmecSurface{T},
                         wmn::AbstractArray{SurfaceFourierData{T}},
                         boozerG::T,
                         boozerI::T;
                        ) where {T <: AbstractFloat, N, C, I}
  res = Array{T}(undef, size(v))
  @batch minbatch=16 for i in eachindex(v, res)
    res[i] = boozer_jacobian(BoozerFromFlux(), v[i], vmec, wmn, boozerG, boozerI)
  end
  return res
end

@doc raw"""
    periodic_boozer_function(vmec::VmecSurface)

Compute the auxillary Boozer transformation data `wmn`, `boozerG`, and `boozerI`,
where `wmn` is the `SurfaceFourierData` spectrum ``\frac{B_\theta}{m} - \frac{B_\phi}{n}``
and `boozerG` is the enclosed toroidal current and `boozerI` is the enclosed poloidal current.

See also: [`BoozerFromFlux`](@ref)
"""
function periodic_boozer_function(vmec::VmecSurface{T};
                                 ) where {T <: AbstractFloat}
  boozerG = vmec.bsubvmn[1].cos
  boozerI = vmec.bsubumn[1].cos
  zeros_T = zeros(T,vmec.mnmax_nyq)
  wmn_cos = Array{T}(undef,vmec.mnmax_nyq)
  wmn_sin = similar(wmn_cos)
  for mn in eachindex(wmn_cos,wmn_sin)
    m = vmec.xm_nyq[mn]
    n = vmec.xn_nyq[mn]
    if m != 0 && n != 0
      wmn_cos[mn] = 0.5*(vmec.bsubumn[mn].cos/m - vmec.bsubvmn[mn].cos/n)
      wmn_sin[mn] = 0.5*(vmec.bsubumn[mn].sin/m - vmec.bsubvmn[mn].sin/n)
   elseif m != 0
      wmn_cos[mn] = vmec.bsubumn[mn].cos/m
    wmn_sin[mn] = vmec.bsubumn[mn].sin/m
    elseif n != 0
      wmn_cos[mn] = -vmec.bsubvmn[mn].cos/n
      wmn_sin[mn] = -vmec.bsubvmn[mn].sin/n
    else
      wmn_cos[mn] = zero(T)
      wmn_sin[mn] = zero(T)
    end
  end
  return StructArray{SurfaceFourierData{T}}((vmec.xm_nyq,vmec.xn_nyq,wmn_cos,zeros_T,wmn_sin,zeros_T)), boozerG, boozerI
end

@doc raw"""
    boozer_grid(vmec::VmecSurface, mBoozer, nBoozer)

Construct a grid of flux coordinates and associated Boozer coordinates
of size (2(2`mBoozer`+1) × 2(2`nBoozer`)+1)).
"""
function boozer_grid(vmec::VmecSurface,
                     mBoozer::Int,
                     nBoozer::Int;
                    ) where {T <: AbstractFloat}
  θ_range = range(0.0,step=2π/(2*(2*mBoozer+1)),length=2*(2*mBoozer+1))
  ζ_range = range(0.0,step=2π/(vmec.nfp*2*(2*nBoozer+1)),length=2(2*nBoozer+1))
  u = MagneticCoordinateGrid(FluxCoordinates,vmec.s*vmec.phi[end]/2π*vmec.signgs,θ_range,ζ_range)
  wmn, boozerG, boozerI = periodic_boozer_function(vmec)
  b = BoozerFromFlux()(u, vmec, wmn, boozerG, boozerI)
  jac = boozer_jacobian(BoozerFromFlux(), u, vmec, wmn, boozerG, boozerI)
  return u, b, jac
end

@doc raw"""
    boozer_mode_indices(mBoozer, nBoozer, nfp)

Construct the list of unique Boozer mode indices for a given `mBoozer`, `nBoozer` and `nfp`.
"""
function boozer_mode_indices(mBoozer::Int,
                             nBoozer::Int,
                             nfp::Int;
                            )
  mnBoozer = nBoozer + 1 + (mBoozer-1)*(1 + 2*nBoozer)
  modes = Vector{SVector{2,Int}}(undef,mnBoozer)
  i = 1
  for n = 0:nfp:nfp*nBoozer
    modes[i] = SVector{2,Int}(0,n)
    i += 1
  end
  for m = 1:mBoozer-1
    for n = -nfp*nBoozer:nfp:nfp*nBoozer
      modes[i] = SVector{2,Int}(m,n)
      i += 1
    end
  end
  return modes, mnBoozer
end

@doc raw"""
    boozer_cosine_spectrum(field::Symbol, vmec::VmecSurface, mBoozer, nBoozer)
    boozer_cosine_spectrum(field::Symbol, surfaces, vmec::Vmec, mBoozer, nBoozer)

Compute the Boozer mode amplitudes for a VMEC field `field`
on a `VmecSurface` or list of surfaces from a VMEC equilibrium
defined by a cosine series representation using `mBoozer` × `nBoozer` modes.
"""
function boozer_cosine_spectrum(field::Symbol,
                                vmec::VmecSurface{T},
                                mBoozer::Int,
                                nBoozer::Int;
                               ) where {T}
  modes, mnBoozer = boozer_mode_indices(mBoozer, nBoozer, vmec.nfp)
  spectrum = !vmec.lasym ? Vector{SVector{3, T}}(undef, mnBoozer) : Vector{SVector{4, T}}(undef, mnBoozer)
  scaleFactor = 2.0/(2 * (2 * mBoozer + 1) * 2 * (2 * nBoozer + 1))
  #Set up the coordinates, field data and Jacobian of the transformation
  u, b, jac = boozer_grid(vmec, mBoozer, nBoozer)
  data = Matrix{T}(undef, size(u))
  for i in eachindex(u, jac, data)
    data[i] = inverseTransform(u[i], getfield(vmec, field)) * jac[i]
  end
  @batch minbatch=128 for i in eachindex(modes)
    spectrum[i] = SVector{3,T}(modes[i][1], modes[i][2],
                          cosineTransform(modes[i][1], modes[i][2], b, data)*scaleFactor)
  end
  spectrum[1] = SVector(0, 0, spectrum[1][3]*0.5)
  return spectrum
end

function boozer_cosine_spectrum(field::Symbol,
                                surfaces::AbstractVector{T},
                                vmec::Vmec{T},
                                mBoozer::Int,
                                nBoozer::Int;
                               ) where {T <: AbstractFloat}
  spectrumData = Vector{Vector{SVector{3, T}}}(undef, length(surfaces))
  for i in eachindex(surfaces, spectrumData)
    vmecSurface = VmecSurface(surfaces[i], vmec)
    spectrumData[i] = boozer_cosine_spectrum(field, vmecSurface, mBoozer, nBoozer)
  end
  return spectrumData
end

@doc raw"""
    quasisymmetry_deviation(mQS, nQS, mBoozer, nBoozer, vmec::VmecSurface)
    quasisymmetry_deviation(mQS, nQS, mBoozer, nBoozer, vmecSurfaces::Vector{VmecSurface})

Compute the quasisymmetry deviation measure of a `VmecSurface`:

``QS_{deviation} = \frac{\sum\limits_{m', n'} B_{m',n'}^2}{B_{0,0}^2}, \text{with } m' \neq \alpha m_{QS},\,n' \neq \alpha n_{QS}``

where ``\alpha \in \mathbb{Z}``.  If a vector of `VmecSurface` objects are supplied,
the quasisymmetry deviation will be computed at each surface.
"""
function quasisymmetry_deviation(mQS::Int,
                                 nQS::Int,
                                 mBoozer::Int,
                                 nBoozer::Int,
                                 vmec::VmecSurface{T};
                                ) where {T <: AbstractFloat}
  modB = boozer_cosine_spectrum(:bmn, vmec, mBoozer, nBoozer)
  polModes = map(i->i[1], modB)
  torModes = map(i->i[2], modB)
  denom = abs(modB[1][3])
  asymPower = sum(map(i->i[3]^2, modB[2:end]))
  m = mQS
  n = nQS
  while m <= (mBoozer-1) && n <= vmec.nfp*nBoozer
    index = findfirst((polModes.==m).*(torModes.==n).==1)
    asymPower -= modB[index][3]^2
    m += mQS
    n += nQS
  end
  return asymPower > zero(typeof(asymPower)) ? sqrt(asymPower)/denom : zero(typeof(asymPower))
end

function quasisymmetry_deviation(mQS::Int,
                                 nQS::Int,
                                 mBoozer::Int,
                                 nBoozer::Int,
                                 vmecSurfaces::AbstractVector{VmecSurface{T}};
                                ) where {T}
  qsValues = Vector{T}(undef,length(vmecSurfaces))
  for i in eachindex(qsValues,vmecSurfaces)
    qsValues[i] = quasisymmetry_deviation(mQS,nQS,mBoozer,nBoozer,vmecSurfaces[i])
  end
  return qsValues
end
