
struct CylindricalFromVmec <: Transformation; end
struct CartesianFromVmec <: Transformation; end
struct VmecFromFlux <: Transformation; end
struct FluxFromVmec <: Transformation; end
struct VmecFromPest <: Transformation; end
struct PestFromVmec <: Transformation; end
struct BoozerFromVmec <: Transformation; end
struct BoozerFromFlux <: Transformation; end
#struct Derivatives end

function (::VmecFromFlux)(x::FluxCoordinates,
                          vmec::VmecSurface;
                         )
  y = FluxCoordinates(x.ψ,x.θ,-x.ζ)
  s = x.ψ*2π/vmec.phi[end]*vmec.signgs
  θ = thetaVmec(y,vmec)
  ζ = -x.ζ
  return VmecCoordinates(s,θ,ζ)
end

function (::CylindricalFromVmec)(x::VmecCoordinates,
                                 vmec::VmecSurface;
                                )
  R = inverseTransform(x,vmec.rmn)
  Z = inverseTransform(x,vmec.zmn)
  ϕ = x.ζ
  return Cylindrical(R,ϕ,Z)
end

function (::CylindricalFromFlux)(x::FluxCoordinates,
                                 vmec::VmecSurface;
                                )
  v = VmecFromFlux()(x,vmec)
  return CylindricalFromVmec()(v,vmec)
end

#=
function (::CylindricalFromVmec)(x::VmecCoordinates,
                                 vmec::VmecSurface,
                                 ::Derivatives;
                                )
  T = typeof(x.θ)
  d = derivatives(x, vmec)
  return CylindricalFromVmec()(x,vmec), @SMatrix [d[1,1] d[1,2] d[1,3];
                                                  zero(T) zero(T) one(T);
                                                  d[3,1] d[3,2] d[3,3]]
end
=#

function (::CartesianFromVmec)(x::VmecCoordinates,
                               vmec::VmecSurface;
                              )
  c = CylindricalFromVmec()(x,vmec)
  return CartesianFromCylindrical()(c)
end

function (::CartesianFromFlux)(x::FluxCoordinates,
                               vmec::VmecSurface;
                              )
  v = VmecFromFlux(x,vmec)
  return CartesianFromVmec(v,vmec)
end

function (::FluxFromPest)(x::PestCoordinates,
                          vmec::VmecSurface;
                         )
  # VMEC is a left handed coordinate system and a positive rotational transform
  # implies the theta coordinate increases with positive toroidal angle (CCW direction).
  # FluxCoordinates and PestCoordinates are right handed coordinate systems, so to match
  # that ∇ψ > 0 and ∇θ > 0 at (θ,ζ) = (0,0), increasing toroidal angle is negative of the
  # cylindrical toroidal angle and increases CW.  This means the rotational transform
  # in these coordinates is the negative of the rotational transform in VMEC
  return FluxCoordinates(x.ψ,x.α-vmec.iota[1]*x.ζ,x.ζ)
end

function (::VmecFromPest)(x::PestCoordinates,
                          vmec::VmecSurface;
                         )
  y = FluxFromPest()(x,vmec)
  return VmecFromFlux()(y,vmec)
end

function (::FluxFromVmec)(x::VmecCoordinates,
                        vmec::VmecSurface;
                       )
  ψ = vmec.phi[1]/(2π)*vmec.signgs
  θ = x.θ + inverseTransform(x, vmec.λmn)
  ζ = -x.ζ
  return FluxCoordinates(ψ, θ, ζ)
end

function (::PestFromVmec)(x::VmecCoordinates,
                          vmec::VmecSurface;
                         )
  y = FluxFromVmec()(x, vmec)
  α = y.θ + vmec.iota[1] * y.ζ
  return PestCoordinates(y.ψ, α, y.ζ)
end

function transform_deriv(::CylindricalFromFlux,
                         x::FluxCoordinates,
                         vmec::VmecSurface;
                        )
  v = VmecFromFlux()(x,vmec)
  dV = derivatives(v,vmec)
  dsdψ = 2π/vmec.phi[end]*vmec.signgs
  dΛdθv = dV[2,2]
  dΛdζv = dV[2,3]

  dRdψ = dV[1,1]*dsdψ
  dZdψ = dV[3,1]*dsdψ
  dϕdψ = zero(typeof(x.θ))

  dRdθ = dV[1,2]/(1+dΛdθv)
  dZdθ = dV[3,2]/(1+dΛdθv)
  dϕdθ = zero(typeof(x.θ))

  dRdζ = -dV[1,3] + dV[1,2]*dΛdζv/(1+dΛdθv)
  dZdζ = -dV[3,3] + dV[3,2]*dΛdζv/(1+dΛdθv)
  dϕdζ = one(typeof(x.θ))
  return @SMatrix [dRdψ dRdθ dRdζ;
                   dϕdψ dϕdθ dϕdζ;
                   dZdψ dZdθ dZdζ]
end

function transform_deriv(::CylindricalFromVmec,
                         x::VmecCoordinates,
                         vmec::VmecSurface;
                        )
  dRds = inverseTransform(x,vmec.rmn;deriv=:ds)
  dZds = inverseTransform(x,vmec.zmn;deriv=:ds)
  dϕds = zero(typeof(x.θ))

  dRdθ = inverseTransform(x,vmec.rmn;deriv=:dθ)
  dZdθ = inverseTransform(x,vmec.zmn;deriv=:dθ)
  dϕdθ = zero(typeof(x.θ))

  dRdζ = inverseTransform(x,vmec.rmn;deriv=:dζ)
  dZdζ = inverseTransform(x,vmec.zmn;deriv=:dζ)
  dϕdζ = one(typeof(x.θ))
  return @SMatrix [dRds dRdθ dRdζ;
                   dϕds dϕdθ dϕdζ;
                   dZds dZdθ dZdζ]
end

function derivatives(x::VmecCoordinates, vmec::VmecSurface)
  dRds = inverseTransform(x,vmec.rmn;deriv=:ds)
  dZds = inverseTransform(x,vmec.zmn;deriv=:ds)
  dΛds = inverseTransform(x,vmec.λmn;deriv=:ds)

  dRdθ = inverseTransform(x,vmec.rmn;deriv=:dθ)
  dZdθ = inverseTransform(x,vmec.zmn;deriv=:dθ)
  dΛdθ = inverseTransform(x,vmec.λmn;deriv=:dθ)

  dRdζ = inverseTransform(x,vmec.rmn;deriv=:dζ)
  dZdζ = inverseTransform(x,vmec.zmn;deriv=:dζ)
  dΛdζ = inverseTransform(x,vmec.λmn;deriv=:dζ)
  return @SMatrix [dRds dRdθ dRdζ;
                   dΛds dΛdθ dΛdζ;
                   dZds dZdθ dZdζ]
end

function derivatives(x::VmecCoordinates, vmec::VmecSurface, field::Symbol)
  ds = inverseTransform(x,getfield(vmec,field);deriv=:ds)
  dθ = inverseTransform(x,getfield(vmec,field);deriv=:dθ)
  dζ = inverseTransform(x,getfield(vmec,field);deriv=:dζ)
  return @SVector [ds, dθ, dζ]
end

#function transform_deriv(::CartesianFromVmec, x::VmecCoordinates, vmec::VmecSurface)
#  y = transform_deriv(CylindricalFromVmec(),x,vmec)
#  return transform_deriv(CartesianFromCylindrical(),y)
#end

function PlasmaEquilibriumToolkit.basis_vectors(::Covariant,
                                                ::CartesianFromVmec,
                                                x::VmecCoordinates,
                                                vmec::VmecSurface
                                               )
  c = CylindricalFromVmec()(x,vmec)
  d = transform_deriv(CylindricalFromVmec(),x,vmec)
  sζ, cζ = sincos(x.ζ)
  dXds_x = d[1,1]*cζ
  dXds_y = d[1,1]*sζ
  dXds_z = d[3,1]

  dXdθ_x = d[1,2]*cζ
  dXdθ_y = d[1,2]*sζ
  dXdθ_z = d[3,2]

  dXdζ_x = d[1,3]*cζ - c.r*sζ
  dXdζ_y = d[1,3]*sζ + c.r*cζ
  dXdζ_z = d[3,3]

  return @SMatrix [dXds_x dXdθ_x dXdζ_x;
                   dXds_y dXdθ_y dXdζ_y;
                   dXds_z dXdθ_z dXdζ_z]
end

function PlasmaEquilibriumToolkit.basis_vectors(::Contravariant,
                                                ::CartesianFromVmec,
                                                x::VmecCoordinates,
                                                vmec::VmecSurface;
                                               )
  g = inverseTransform(x,vmec.gmn)
  return transform_basis(ContravariantFromCovariant(),basis_vectors(Covariant(),CartesianFromVmec(),x,vmec),g)
end

function PlasmaEquilibriumToolkit.basis_vectors(::Covariant,
                                                ::CartesianFromPest,
                                                x::PestCoordinates,
                                                vmec::VmecSurface;
                                               )
  contravariantPestBasis = basis_vectors(Contravariant(),CartesianFromPest(),x,vmec)
  return transform_basis(CovariantFromContravariant(),contravariantPestBasis)
end

function PlasmaEquilibriumToolkit.basis_vectors(::Contravariant,
                                                ::CartesianFromPest,
                                                x::PestCoordinates,
                                                vmec::VmecSurface;
                                               )
  v = VmecFromPest()(x,vmec)
  contravariantVmecBasis = basis_vectors(Contravariant(),CartesianFromVmec(),v,vmec)
  return transform_basis(PestFromVmec(),v,contravariantVmecBasis,vmec)
end

function PlasmaEquilibriumToolkit.basis_vectors(::Covariant,
                                                ::CartesianFromFlux,
                                                x::FluxCoordinates,
                                                vmec::VmecSurface;
                                               )
  contravariantFluxBasis = basis_vectors(Contravariant(),CartesianFromFlux(),x,vmec)
  return transform_basis(CovariantFromContravariant(),contravariantFluxBasis)
end

function PlasmaEquilibriumToolkit.basis_vectors(::Contravariant,
                                                ::CartesianFromFlux,
                                                x::FluxCoordinates,
                                                vmec::VmecSurface;
                                               )
  v = VmecFromFlux()(x,vmec)
  contravariantVmecBasis = basis_vectors(Contravariant(),CartesianFromVmec(),v,vmec)
  return transform_basis(FluxFromVmec(),v,contravariantVmecBasis,vmec)
end

# The transform_basis functions need to take into account the definitions of rotational
# transform in the respective coordinate systems
# αₚ = θₚ - ιₚζₚ = θᵥ + λ(s,θᵥ,ϕᵥ) - (-ιᵥ)(-ϕᵥ)
# => ∇αₚ = ∇θᵥ + dλ/ds ∇s + dλ/dθᵥ ∇θᵥ + dλ/dϕᵥ ∇ϕᵥ - dιᵥ/ds ϕᵥ ∇s - ιᵥ ∇ϕᵥ
# => ∇αₚ = (dλ/ds - dιᵥ/ds ϕᵥ)∇s+ (1 + dλ/dθᵥ)∇θᵥ + (dλ/dϕᵥ - ιᵥ)∇ϕᵥ
function PlasmaEquilibriumToolkit.transform_basis(::PestFromVmec,
                                                  v::VmecCoordinates,
                                                  e::BasisVectors,
                                                  vmec::VmecSurface;
                                                 )
  d = derivatives(v,vmec,:λmn)
  Φ = vmec.phi[end]/(2π)*vmec.signgs
  ∇ψ = Φ*e[:,1]
  ∇ζ = -e[:,3]
  ∇α = sign(Φ)*(e[:,1]*(d[1] - vmec.iota[2]*v.ζ)
                + e[:,2]*(d[2] + 1) +
                e[:,3]*(d[3] - vmec.iota[1]))
  return hcat(∇ψ, ∇α, ∇ζ)
end

# => ∇θ = ∇θᵥ + dλ/ds ∇s + dλ/dθᵥ ∇θᵥ + dλ/dϕᵥ ∇ϕᵥ
function PlasmaEquilibriumToolkit.transform_basis(::FluxFromVmec,
                                                  v::VmecCoordinates,
                                                  e::BasisVectors,
                                                  vmec::VmecSurface;
                                                 )
  d = derivatives(v,vmec,:λmn)
  Φ = vmec.phi[end]/(2π)*vmec.signgs
  ∇ψ = Φ*e[:,1]
  ∇ζ = -e[:,3]
  ∇θ = sign(Φ)*(e[:,1]*d[1] + e[:,2]*(1 + d[2]) + e[:,3]*d[3])
  return hcat(∇ψ, ∇θ, ∇ζ)
end

# Transform coordinates to the GENE coordinate system
function PlasmaEquilibriumToolkit.transform_basis(::GeneFromPest,
                                                  p::PestCoordinates,
                                                  e::BasisVectors,
                                                  vmec::VmecSurface;
                                                 )
  Aminor = vmec.Aminor_p
  Φ = vmec.phi[end]*vmec.signgs
  x = sqrt(p.ψ*2π/Φ)
  ∇x = Aminor*π/(Φ*x)*e[:,1]
  # Multiple by sign(Φ) to ensure coordinate system is left handed
  ∇y = sign(Φ)*Aminor*x*e[:,2]
  ∇z = Aminor*(e[:,2] - p.ζ*vmec.iota[2]*2π/Φ*e[:,1] - vmec.iota[1]*e[:,3])
  return hcat(∇x, ∇y, ∇z)
end

# Replicate a GIST file
function geneGeometryCoefficients(::GeneFromPest,
                                  p::PestCoordinates,
                                  vmec::VmecSurface;
                                 )
  Aminor = vmec.Aminor_p
  Ba = abs(vmec.phi[end])/(π*Aminor^2)
  v = VmecFromPest()(p,vmec)
  vmecContraBasis = basis_vectors(Contravariant(), CartesianFromVmec(), v, vmec)
  pestContraBasis = transform_basis(PestFromVmec(), v, vmecContraBasis,vmec)
  geneContraBasis = transform_basis(GeneFromPest(), p, pestContraBasis,vmec)
  geneCoBasis = transform_basis(CovariantFromContravariant(), geneContraBasis)
  gB = grad_B(v, vmecContraBasis, vmec)
  geneMetric = metric(geneContraBasis)
  BnormMag = norm(cross(geneContraBasis[:,1], geneContraBasis[:,2]))
  # The Jacobian should always be negative, but as only the integral measure
  # is required, that must be positive negative
  jac = abs(1.0/dot(geneContraBasis[:,1], cross(geneContraBasis[:,2], geneContraBasis[:,3])))
  K1, K2 = Aminor/Ba.*grad_B_projection(geneContraBasis,gB)
  dBdZ = Aminor/Ba*dot(gB, geneCoBasis[:,3])
  # The GENE K1 component is the negative of what is computed, see the GIST documentation note
  return geneMetric, BnormMag, abs(jac), -K1, K2, dBdZ
end

function geneGeometryCoefficients(::GeneFromPest,
                                  p::StructArray{PestCoordinates{T,T}},
                                  vmec::VmecSurface
                                 ) where {T}
  geneMetric = Array{SVector{6,Float64}}(undef, size(p))
  K1 = Array{Float64}(undef, size(p))
  K2 = similar(K1)
  dBdZ = similar(K1)
  BnormMag = similar(K1)
  absJac = similar(K1)
  @batch for i in eachindex(p)
    geneMetric[i], BnormMag[i], absJac[i], K1[i], K2[i], dBdZ[i] = geneGeometryCoefficients(GeneFromPest(),p[i],vmec)
  end
  return geneMetric, BnormMag, absJac, K1, K2, dBdZ
end

# Compute the normalized global magnetic shear
function shat(vmec::VmecSurface)
  return -2*vmec.s*vmec.iota[2]/vmec.iota[1]
end
