function evaluate_exact(x::VmecCoordinates, valmn)
  return inverseTransform(x, valmn)
end

function evaluate_exact_deriv(x::VmecCoordinates, valmn)
  ds = inverseTransform(x, valmn, deriv=:ds)
  dθ = inverseTransform(x, valmn, deriv=:dθ)
  dζ = inverseTransform(x, valmn, deriv=:dζ)
  return (ds, dθ, dζ)
end

function make_surface_interpolation(valmn, 
                                    nfp::Int, 
                                    θres::Int,
                                    ζres::Int)

  θs = range(0, length = θres, step = 2π/θres)
  ζs = range(0, length = ζres, step = 2π/nfp/ζres)
  knots = (θs, ζs)
  itp_types = (BSpline(Cubic(Periodic(OnCell()))),
                 BSpline(Cubic(Periodic(OnCell()))))
  itp = (f) -> scale(interpolate(f, itp_types), knots...)
  extp = (f) -> extrapolate(itp(f), (Periodic(), Periodic()))
  field = [inverseTransform(VmecCoordinates(0.0, θ, ζ), valmn) 
           for ζ in ζs for θ in θs]
  field = reshape(field, (θres, ζres))
  deriv = [inverseTransform(VmecCoordinates(0.0, θ, ζ), valmn, deriv=:ds) 
           for ζ in ζs for θ in θs]
  deriv = reshape(deriv, (θres, ζres))
  return (extp(field), extp(deriv))
end                                          

function vmec_b(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.bmn)
  end
  if vmec.b == nothing
    (vmec.b, vmec.dbds) = make_surface_interpolation(vmec.bmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.b.itp)[1] < minθres || size(vmec.b.itp)[2] < minζres
    (vmec.b, vmec.dbds) = make_surface_interpolation(vmec.bmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.b(x.θ, x.ζ)
end

function vmec_r(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.rmn)
  end
  if vmec.r == nothing
    (vmec.r, vmec.drds) = make_surface_interpolation(vmec.rmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.r.itp)[1] < minθres || size(vmec.r.itp)[2] < minζres
    (vmec.r, vmec.drds) = make_surface_interpolation(vmec.rmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.r(x.θ, x.ζ)
end

function vmec_z(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.zmn)
  end
  if vmec.z == nothing
    (vmec.z, vmec.dzds) = make_surface_interpolation(vmec.zmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.z.itp)[1] < minθres || size(vmec.z.itp)[2] < minζres
    (vmec.z, vmec.dzds) = make_surface_interpolation(vmec.zmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.z(x.θ, x.ζ)
end

function vmec_λ(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.λmn)
  end
  if vmec.λ == nothing
    (vmec.λ, vmec.dλds) = make_surface_interpolation(vmec.λmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.λ.itp)[1] < minθres || size(vmec.λ.itp)[2] < minζres
    (vmec.λ, vmec.dλds) = make_surface_interpolation(vmec.λmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.λ(x.θ, x.ζ)
end

function vmec_g(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.gmn)
  end
  if vmec.g == nothing
    (vmec.g, vmec.dgds) = make_surface_interpolation(vmec.gmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.g.itp)[1] < minθres || size(vmec.g.itp)[2] < minζres
    (vmec.g, vmec.dgds) = make_surface_interpolation(vmec.gmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.g(x.θ, x.ζ)
end

function vmec_bsubs(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.bsubsmn)
  end
  if vmec.bsubs == nothing
    (vmec.bsubs, vmec.dbsubsds) = make_surface_interpolation(vmec.bsubsmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsubs.itp)[1] < minθres || size(vmec.bsubs.itp)[2] < minζres
    (vmec.bsubs, vmec.dbsubsds) = make_surface_interpolation(vmec.bsubsmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.bsubs(x.θ, x.ζ)
end

function vmec_bsubu(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.bsubumn)
  end
  if vmec.bsubu == nothing
    (vmec.bsubu, vmec.dbsubuds) = make_surface_interpolation(vmec.bsubumn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsubu.itp)[1] < minθres || size(vmec.bsubu.itp)[2] < minζres
    (vmec.bsubu, vmec.dbsubuds) = make_surface_interpolation(vmec.bsubumn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.bsubu(x.θ, x.ζ)
end

function vmec_bsubv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.bsubvmn)
  end
  if vmec.bsubv == nothing
    (vmec.bsubv, vmec.dbsubvds) = make_surface_interpolation(vmec.bsubvmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsubv.itp)[1] < minθres || size(vmec.bsubv.itp)[2] < minζres
    (vmec.bsubv, vmec.dbsubvds) = make_surface_interpolation(vmec.bsubvmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.bsubv(x.θ, x.ζ)
end

function vmec_bsupu(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.bsupumn)
  end
  if vmec.bsupu == nothing
    (vmec.bsupu, vmec.dbsupuds) = make_surface_interpolation(vmec.bsupumn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsupu.itp)[1] < minθres || size(vmec.bsupu.itp)[2] < minζres
    (vmec.bsupu, vmec.dbsupuds) = make_surface_interpolation(vmec.bsupumn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.bsupu(x.θ, x.ζ)
end

function vmec_bsupv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.bsupvmn)
  end
  if vmec.bsupv == nothing
    (vmec.bsupv, vmec.dbsupvds) = make_surface_interpolation(vmec.bsupvmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsupv.itp)[1] < minθres || size(vmec.bsupv.itp)[2] < minζres
    (vmec.bsupv, vmec.dbsupvds) = make_surface_interpolation(vmec.bsupvmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.bsupv(x.θ, x.ζ)
end

function vmec_curru(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.currumn)
  end
  if vmec.curru == nothing
    (vmec.curru, vmec.dcurruds) = make_surface_interpolation(vmec.currumn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.curru.itp)[1] < minθres || size(vmec.curru.itp)[2] < minζres
    (vmec.curru, vmec.dcurruds) = make_surface_interpolation(vmec.currumn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.curru(x.θ, x.ζ)
end

function vmec_currv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact(x, vmec.currvmn)
  end
  if vmec.currv == nothing
    (vmec.currv, vmec.dcurrvds) = make_surface_interpolation(vmec.currvmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.currv.itp)[1] < minθres || size(vmec.currv.itp)[2] < minζres
    (vmec.currv, vmec.dcurrvds) = make_surface_interpolation(vmec.currvmn,
                            vmec.nfp, minθres, minζres)
  end
  return vmec.currv(x.θ, x.ζ)
end

function vmec_b_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.bmn)
  end
  if vmec.b == nothing
    (vmec.b, vmec.dbds) = make_surface_interpolation(vmec.bmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.b.itp)[1] < minθres || size(vmec.b.itp)[2] < minζres
    (vmec.b, vmec.dbds) = make_surface_interpolation(vmec.bmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dbds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.b, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_r_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.rmn)
  end
  if vmec.r == nothing
    (vmec.r, vmec.drds) = make_surface_interpolation(vmec.rmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.r.itp)[1] < minθres || size(vmec.r.itp)[2] < minζres
    (vmec.r, vmec.drds) = make_surface_interpolation(vmec.rmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.drds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.r, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_z_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.zmn)
  end
  if vmec.z == nothing
    (vmec.z, vmec.dzds) = make_surface_interpolation(vmec.zmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.z.itp)[1] < minθres || size(vmec.z.itp)[2] < minζres
    (vmec.z, vmec.dzds) = make_surface_interpolation(vmec.zmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dzds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.z, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_λ_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.λmn)
  end
  if vmec.λ == nothing
    (vmec.λ, vmec.dλds) = make_surface_interpolation(vmec.λmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.λ.itp)[1] < minθres || size(vmec.λ.itp)[2] < minζres
    (vmec.λ, vmec.dλds) = make_surface_interpolation(vmec.λmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dλds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.λ, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_g_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.gmn)
  end
  if vmec.g == nothing
    (vmec.g, vmec.dgds) = make_surface_interpolation(vmec.gmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.g.itp)[1] < minθres || size(vmec.g.itp)[2] < minζres
    (vmec.g, vmec.dgds) = make_surface_interpolation(vmec.gmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dgds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.g, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_bsubs_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.bsubsmn)
  end
  if vmec.bsubs == nothing
    (vmec.bsubs, vmec.dbsubsds) = make_surface_interpolation(vmec.bsubsmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsubs.itp)[1] < minθres || size(vmec.bsubs.itp)[2] < minζres
    (vmec.bsubs, vmec.dbsubsds) = make_surface_interpolation(vmec.bsubsmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dbsubsds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.bsubs, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_bsubu_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.bsubumn)
  end
  if vmec.bsubu == nothing
    (vmec.bsubu, vmec.dbsubuds) = make_surface_interpolation(vmec.bsubumn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsubu.itp)[1] < minθres || size(vmec.bsubu.itp)[2] < minζres
    (vmec.bsubu, vmec.dbsubuds) = make_surface_interpolation(vmec.bsubumn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dbsubuds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.bsubu, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_bsubv_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.bsubvmn)
  end
  if vmec.bsubv == nothing
    (vmec.bsubv, vmec.dbsubvds) = make_surface_interpolation(vmec.bsubvmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsubv.itp)[1] < minθres || size(vmec.bsubv.itp)[2] < minζres
    (vmec.bsubv, vmec.dbsubvds) = make_surface_interpolation(vmec.bsubvmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dbsubvds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.bsubv, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_bsupu_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.bsupumn)
  end
  if vmec.bsupu == nothing
    (vmec.bsupu, vmec.dbsupuds) = make_surface_interpolation(vmec.bsupumn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsupu.itp)[1] < minθres || size(vmec.bsupu.itp)[2] < minζres
    (vmec.bsupu, vmec.dbsupuds) = make_surface_interpolation(vmec.bsupumn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dbsupuds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.bsupu, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_bsupv_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.bsupvmn)
  end
  if vmec.bsupv == nothing
    (vmec.bsupv, vmec.dbsupvds) = make_surface_interpolation(vmec.bsupvmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.bsupv.itp)[1] < minθres || size(vmec.bsupv.itp)[2] < minζres
    (vmec.bsupv, vmec.dbsupvds) = make_surface_interpolation(vmec.bsupvmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dbsupvds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.bsupv, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_curru_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.currumn)
  end
  if vmec.curru == nothing
    (vmec.curru, vmec.dcurruds) = make_surface_interpolation(vmec.currumn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.curru.itp)[1] < minθres || size(vmec.curru.itp)[2] < minζres
    (vmec.curru, vmec.dcurruds) = make_surface_interpolation(vmec.currumn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dcurruds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.curru, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

function vmec_currv_deriv(x::VmecCoordinates,
                       vmec::VmecSurface{T};
                       minθres = 128,
                       minζres = 128,
                       exact = false) where {T}
  if exact
    return evaluate_exact_deriv(x, vmec.currvmn)
  end
  if vmec.currv == nothing
    (vmec.currv, vmec.dcurrvds) = make_surface_interpolation(vmec.currvmn,
                            vmec.nfp, minθres, minζres)
  elseif size(vmec.currv.itp)[1] < minθres || size(vmec.currv.itp)[2] < minζres
    (vmec.currv, vmec.dcurrvds) = make_surface_interpolation(vmec.currvmn,
                            vmec.nfp, minθres, minζres)
  end
  ds = vmec.dcurrvds(x.θ, x.ζ)
  (dθ, dζ) = Interpolations.gradient(vmec.currv, x.θ, x.ζ)
  return (ds, dθ, dζ)
end

